from os.path import dirname
from pathlib import Path
import sys
from trilogy_public_models.models import LazyEnvironment, ModelOutput, ModelDict
from collections import UserDict


def discover_models(models):
    import os

    base = dirname(__file__)

    for root, dirs, files in os.walk(base):
        for f in files:
            if f == "entrypoint.preql":
                relative = Path(root).relative_to(base)
                path = ".".join(relative.parts)
                models[path] = None
                query_array = []
                lazy_env = LazyEnvironment(
                    load_path=Path(root) / f,
                    working_path=Path(root),
                    setup_queries=query_array,
                )
                output = ModelOutput(environment=lazy_env, setup=lazy_env.setup_queries)
                sys.modules["trilogy_public_models." + path] = output.environment
                models[path] = output


data_models: UserDict["str", ModelOutput] = ModelDict()
discover_models(data_models)



from os import listdir
from os.path import dirname, join

from trilogy import Environment
from trilogy.constants import ENV_CACHE_NAME
from trilogy.parser import parse
from trilogy.authoring import RawSQLStatement, SelectStatement, PersistStatement


def parse_initial_models(
    fpath: str,
) -> tuple[Environment, list[RawSQLStatement | SelectStatement | PersistStatement]]:
    parent_folder = dirname(fpath)
    files = listdir(dirname(fpath))
    cache_path = join(parent_folder, ENV_CACHE_NAME)
    # 2024-03-31 - disable due to issue with pydantic parsing from cache wrong
    # for file in files:
    #     if file == ENV_CACHE_NAME:
    #         env = Environment.from_cache(join(parent_folder, file))
    #         if env:
    #             return env
    for file in files:
        if file.endswith("entrypoint.preql"):
            with open(join(parent_folder, file), "r", encoding="utf-8") as f:
                contents = f.read()
                env = Environment(working_path=parent_folder)
                environment, statements = parse(contents, environment=env)
                env.to_cache(cache_path)
                return environment, statements

    raise ValueError(f"Missing entrypoint.preql in {fpath}")



from trilogy import Environment, Dialects
from trilogy.constants import DEFAULT_NAMESPACE
from trilogy.authoring import (
    Concept,
    ConceptRef,
    SelectStatement,
)
from trilogy.core.models.datasource import Datasource
from trilogy.core.statements.execute import ProcessedShowStatement
from trilogy.core.processing.concept_strategies_v3 import search_concepts, History
from trilogy.executor import Executor
from trilogy.parser import parse_text
from trilogy.core.internal import INTERNAL_NAMESPACE
from trilogy.core.env_processor import generate_graph
from trilogy.core.models.build import Factory, BuildConcept
from trilogy.core.models.build_environment import BuildEnvironment


def safe_address(input: Concept | ConceptRef):
    if input.namespace == DEFAULT_NAMESPACE:
        return input.name
    return input.address


def validate_dataset(
    dataset: Datasource, environment: Environment, executor: Executor, dry_run_client
):

    validation_query = (
        "SELECT\n "
        + ",\n\t".join([safe_address(x) for x in dataset.concepts])
        + " LIMIT 0;"
    )

    try:
        _, parsed = parse_text(validation_query, environment)
        processed: list[SelectStatement] = [
            x for x in parsed if isinstance(x, SelectStatement)
        ]
        sql = executor.generator.generate_queries(environment, processed)
    except Exception as e:
        print(validation_query)
        raise e
    for statement in sql:
        if isinstance(statement, ProcessedShowStatement):
            continue
        compiled_sql = ""
        # Start the query, passing in the extra configuration.
        try:
            # for UI execution, cap the limit
            compiled_sql = executor.generator.compile_statement(statement)

            # TODO: implement this
            # rs = executor.engine.dry_run(compiled_sql)
            if executor.dialect == Dialects.BIGQUERY:
                # use a dry run to save costs
                # TODO: move ths into executor so we don't need this import
                from google.cloud.bigquery import QueryJobConfig

                job_config = QueryJobConfig(dry_run=True, use_query_cache=False)
                query_job = dry_run_client.query(
                    compiled_sql, job_config=job_config
                )  # Make an API request.
                print(
                    "This query will process {} bytes.".format(
                        query_job.total_bytes_processed
                    )
                )
            elif executor.dialect == Dialects.DUCK_DB:
                # use a dry run to save costs
                query_job = executor.execute_raw_sql(compiled_sql)
            else:
                raise NotImplementedError(
                    f"Validation not implemented for {executor.dialect}"
                )
        except Exception as e:
            print("Failed validation on:")
            print(validation_query)
            print(compiled_sql)
            raise e


def validate_datasource_grain(datasource):
    # TODO: check that count(*) at datasource grain = 1
    # CON: this requires running an real query
    pass


def validate_concept(concept: BuildConcept, history: History, env, graph):
    if concept.namespace == INTERNAL_NAMESPACE or INTERNAL_NAMESPACE in concept.address:
        return
    search_concepts([concept], history=history, environment=env, depth=0, g=graph)


def validate_model(model: Environment, executor: Executor, dry_run_client):
    for dataset in model.datasources.values():
        validate_dataset(dataset, model, executor, dry_run_client)
    history = History(base_environment=model)
    factory = Factory(model)
    build_model: BuildEnvironment = factory.build(model)
    graph = generate_graph(build_model)
    for concept in build_model.concepts.values():
        validate_concept(concept, history, build_model, graph)



from pathlib import Path
from typing import Any
from pydantic import BaseModel
from trilogy import Environment
import sys
from importlib.machinery import SourceFileLoader
from collections import UserDict
from dataclasses import dataclass
from enum import Enum


class QueryType(Enum):
    SQL = "sql"
    TRILOGY = "trilogy"


@dataclass
class SetupQuery:
    query: Any
    type: QueryType


@dataclass
class ModelOutput:
    environment: Environment
    setup: list[SetupQuery]


class ModelDict(UserDict[str, ModelOutput]):
    def __init__(self):
        super().__init__()
        self.not_exists: set[str] = set()

    def _load(self, item: str, path: str):
        loaded = SourceFileLoader(item, path).load_module()
        output = ModelOutput(environment=loaded.model, setup=loaded.statements)
        self[item] = output
        sys.modules["trilogy_public_models." + item] = output  # type: ignore
        return output

    def __getitem__(self, item: str) -> ModelOutput:
        path = str(Path(__file__).parent / item.replace(".", "/") / "__init__.py")
        if item not in self and item not in self.not_exists:
            return self._load(item, path)
        response = super().__getitem__(item)
        # if the key is set but not loaded yet
        if not response:
            return self._load(item, path)
        return response

    def __setitem__(self, key, value):
        super().__setitem__(key, value)


class LazyEnvironment(Environment):
    """Variant of environment to defer parsing of a path
    until relevant attributes accessed."""

    load_path: Path

    working_path: Path
    setup_queries: list[Any]
    loaded: bool = False

    def __init__(self, **data):
        # skip the Environment ini
        # as it will be called late
        BaseModel.__init__(self, **data)

    @property
    def setup_path(self) -> Path:
        return self.load_path.parent / "setup.preql"

    @property
    def setup_path_sql(self) -> Path:
        return self.load_path.parent / "setup.sql"

    def _load(self):
        if self.loaded:
            return
        from trilogy import parse

        env = Environment(working_path=str(self.working_path))

        with open(self.load_path, "r") as f:
            env, _ = parse(f.read(), env)
        if self.setup_path.exists():
            with open(self.setup_path, "r") as f2:
                env, q = parse(f2.read(), env)
                for q in q:
                    self.setup_queries.append(SetupQuery(q, QueryType.TRILOGY))
        if self.setup_path_sql.exists():
            with open(self.setup_path_sql, "r") as f2:
                self.setup_queries.append(SetupQuery(f2.read(), QueryType.SQL))
        self.loaded = True
        self.datasources = env.datasources
        self.concepts = env.concepts
        self.imports = env.imports
        self.alias_origin_lookup = env.alias_origin_lookup
        self.materialized_concepts = env.materialized_concepts
        self.functions = env.functions
        self.data_types = env.data_types
        self.cte_name_map = env.cte_name_map

    def __getattribute__(self, name):
        if name not in (
            "datasources",
            "concepts",
            "imports",
            "materialized_concepts",
            "functions",
            "datatypes",
            "cte_name_map",
        ) or name.startswith("_"):
            return super().__getattribute__(name)
        if not self.loaded:
            self._load()
        return super().__getattribute__(name)
